import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:buffer/buffer.dart';
import 'package:logging/logging.dart';

import './error.dart';

class ProtocolError extends IOException {}

class SocksConnection {
  static final _logger = Logger('SocksConnection');

  final Socket _socket;
  Socket? _socket2 = null;
  late final String _remote;
  final Duration _readTimeout = Duration(seconds: 5);

  SocksConnection(this._socket) {
    _remote = "${_socket.remoteAddress.host}:${_socket.remotePort}";
    _process();
  }

  @override
  String toString() {
    return "$runtimeType($_remote)";
  }

  void _process() async {
    _logger.info("start $this");
    final stream = _socket.timeout(Duration(minutes: 2));
    try {
      await _doProcess(stream);
    } on TimeoutException {
      // nothing
    } on SocketException {
      // nothing
    } on SocksError catch (e) {
      _socket.add([5, e.index, 0]);
    } catch (e) {
      _logger.severe(e);
    }
    _socket.destroy();
    _logger.info("end $this");
  }

  Future<void> _doProcess(Stream<Uint8List> stream) async {
    final input = ByteDataReader();
    (() async {
      try {
        await for (final data in stream) {
          if (_socket2 != null) {
            _socket2?.add(data);
          } else {
            input.add(data);
          }
        }
      } on TimeoutException {}
      _socket2?.destroy();
      _socket.destroy();
    })();
    await _processMethods(input);
    await _processRequest(input);
  }

  Future<void> _processMethods(ByteDataReader input) async {
    // recv methods
    await input.readAhead(2).timeout(_readTimeout);
    final v = input.readUint8();
    if (v != 5) {
      throw ProtocolError();
    }
    final n = input.readUint8();
    await input.readAhead(n).timeout(_readTimeout);
    input.read(n);
    // send method
    _socket.add([5, 0]);
  }

  Future<void> _processRequest(ByteDataReader input) async {
    await input.readAhead(3).timeout(_readTimeout);
    final v = input.readUint8();
    if (v != 5) {
      throw ProtocolError();
    }
    final cmd = input.readUint8();
    if (cmd != 1) {
      throw SocksError.commandNotSupported;
    }
    input.readUint8();
    final host = await _readAddress(input);
    await input.readAhead(2);
    final port = input.readUint16();
    _socket2 = await _connect(host, port);
    await for (final data in _socket2!) {
      _socket.add(data);
    }
    _socket.destroy();
    await _socket2?.done;
  }

  Future<String> _readAddress(ByteDataReader input) async {
    await input.readAhead(1).timeout(_readTimeout);
    final addrType = input.readUint8();
    switch (addrType) {
      case 1:
        await input.readAhead(4).timeout(_readTimeout);
        return InternetAddress.fromRawAddress(input.read(4)).host;
      case 3:
        await input.readAhead(1).timeout(_readTimeout);
        final size = input.readUint8();
        await input.readAhead(size).timeout(_readTimeout);
        return utf8.decode(input.read(size));
      case 4:
        await input.readAhead(16).timeout(_readTimeout);
        return InternetAddress.fromRawAddress(input.read(16)).host;
      default:
        throw SocksError.generalFailure;
    }
  }

  Future<Socket> _connect(String host, int port) async {
    try {
      final timeout = Duration(seconds: 5);
      final socket = await Socket.connect(host, port, timeout: timeout);
      final w = ByteDataWriter();
      w.write([5, 0, 0]);
      if (socket.remoteAddress.type == InternetAddressType.IPv4) {
        w.writeUint8(1);
      } else {
        w.writeUint8(4);
      }
      w.write(socket.remoteAddress.rawAddress);
      w.writeUint16(socket.remotePort);
      _socket.add(w.toBytes());
      return socket;
    } catch (e) {
      throw SocksError.hostUnreachable;
    }
  }
}
